package jp.deadend.noname.treenote;

import java.util.Stack;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.Bundle;
import android.text.ClipboardManager;
import android.text.Editable;
import android.text.TextWatcher;
import android.text.util.Linkify;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.widget.EditText;
import android.util.SparseIntArray;

public class TextEditor extends Activity implements LinearLayoutKeyDispatcher.Callback {
	private static class EditHistoryItem {
		private final int mStart;
		private final int mCount;
		private final CharSequence mBeforeText;

		EditHistoryItem(int st, int co, CharSequence beforetext) {
			mStart = st;
			mCount = co;
			mBeforeText = beforetext;
		}
	}

	static final String KEY_STRING = "TextEditorString";
	static final String KEY_TITLE = "TextEditorTitle";
	private static final int DIALOG_CONFIRM_EXIT = 0;
	private static final int COMMAND_QUIT		= 0;
	private static final int COMMAND_SELECTALL	= 1;
	private static final int COMMAND_CUT		= 2;
	private static final int COMMAND_COPY		= 3;
	private static final int COMMAND_PASTE		= 4;
	private static final int COMMAND_SAVE		= 5;
	private static final int COMMAND_UNDO		= 6;

	protected EditText mEditText;
	private ClipboardManager mClipboard;
	private boolean isModified;
	private Stack<EditHistoryItem> mEditHistory;
	private boolean isUndoing;

	protected void getIntentExtra() {
		Intent it = getIntent();
		if (it != null) {
			String str = it.getStringExtra(KEY_STRING);
			if (str != null) {
				mEditText.setText(str);
			}
			str = it.getStringExtra(KEY_TITLE);
			if (str != null) {
				setTitle("TreeNote: " + str);
			}
		} else {
			finish();
		}
	}

	private void setShortcutKeys() {
        SparseIntArray shortcuts = new SparseIntArray();
        shortcuts.put(KeyEvent.KEYCODE_A, COMMAND_SELECTALL);
        shortcuts.put(TreeNotePrefs.getShortcut(this, getString(R.string.pref_key_cut)),	COMMAND_CUT);
        shortcuts.put(TreeNotePrefs.getShortcut(this, getString(R.string.pref_key_copy)),	COMMAND_COPY);
        shortcuts.put(TreeNotePrefs.getShortcut(this, getString(R.string.pref_key_paste)),	COMMAND_PASTE);
        shortcuts.put(TreeNotePrefs.getShortcut(this, getString(R.string.pref_key_save)),	COMMAND_SAVE);
        shortcuts.put(TreeNotePrefs.getShortcut(this, getString(R.string.pref_key_undo)), 	COMMAND_UNDO);

        LinearLayoutKeyDispatcher llkd = (LinearLayoutKeyDispatcher)findViewById(R.id.MainLayout);
        llkd.setCallback(this);
        llkd.setModKey(TreeNotePrefs.getModKey(this));
        llkd.setShortcutKeys(shortcuts);
        llkd.reset();
	}

	private void setOnClickCommandToView(int viewId, int command) {
		final int cmd = command;
		findViewById(viewId).setOnClickListener(new View.OnClickListener() {
			@Override
			public void onClick(View v) {
				performCommand(cmd);
			}
		});
	}

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		setContentView(R.layout.editor);
		mEditText = (EditText)findViewById(R.id.EditText01);
		mEditText.setTextSize(Integer.parseInt(TreeNotePrefs.getFontSize(this)));
		mEditText.setAutoLinkMask(Linkify.EMAIL_ADDRESSES|Linkify.MAP_ADDRESSES|Linkify.WEB_URLS);
		mEditText.setMovementMethod(ArrowKeyAndLinkMovementMethod.getInstance());
		mClipboard = (ClipboardManager)getSystemService(CLIPBOARD_SERVICE);
		isModified = false;
		getIntentExtra();
		setShortcutKeys();

		mEditHistory = new Stack<EditHistoryItem>();
		isUndoing = false;

		mEditText.addTextChangedListener(new TextWatcher() {
			@Override
			public void beforeTextChanged(CharSequence s, int start, int count, int after) {
				if (!isUndoing) {
					CharSequence seq = s.subSequence(start, start + count);
					mEditHistory.push(new EditHistoryItem(start, after, seq));
				}
			}
			@Override
			public void onTextChanged(CharSequence s, int start, int before, int count) {}
			@Override
			public void afterTextChanged(Editable s) {
				isModified = true;
			}
		});

		if (TreeNotePrefs.getToolbar(this).equals("None")) {
			findViewById(R.id.ButtonDone).setVisibility(View.GONE);
			findViewById(R.id.ToolbarLayout).setVisibility(View.GONE);
		} else {
			setOnClickCommandToView(R.id.ButtonDone,	COMMAND_SAVE);
			setOnClickCommandToView(R.id.ButtonCut,		COMMAND_CUT);
			setOnClickCommandToView(R.id.ButtonCopy,	COMMAND_COPY);
			setOnClickCommandToView(R.id.ButtonPaste,	COMMAND_PASTE);
			setOnClickCommandToView(R.id.ButtonUndo,	COMMAND_UNDO);
		}
	}

	@Override
	public boolean dispatchKeyEvent(KeyEvent event) {
		if (event.getAction() == KeyEvent.ACTION_DOWN && event.getKeyCode() == KeyEvent.KEYCODE_BACK) {
			performCommand(COMMAND_QUIT);
			return true;
		}

		return super.dispatchKeyEvent(event);
	}

	public void performCommand(int cmd) {
		switch (cmd) {
		case COMMAND_SELECTALL:
			mEditText.setSelection(0, mEditText.length());
			break;
		case COMMAND_CUT:
			EditTextUtils.textCut(mEditText, mClipboard);
			break;
		case COMMAND_COPY:
			EditTextUtils.textCopy(mEditText, mClipboard);
			break;
		case COMMAND_PASTE:
			EditTextUtils.textPaste(mEditText, mClipboard);
			break;
		case COMMAND_SAVE:
			processDone();
			break;
		case COMMAND_UNDO:
			undo();
			break;
		case COMMAND_QUIT:
			if (isModified) {
				showDialog(DIALOG_CONFIRM_EXIT);
			} else {
				finish();
			}
			break;
		}
	}

	@Override
	public void openOptionsMenu() {
		super.openOptionsMenu();
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		super.onCreateOptionsMenu(menu);
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.editor_menu, menu);

		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.MenuEditorSaveAndQuit:
			performCommand(COMMAND_SAVE);
			break;
		case R.id.MenuEditorCut:
			performCommand(COMMAND_CUT);
			break;
		case R.id.MenuEditorCopy:
			performCommand(COMMAND_COPY);
			break;
		case R.id.MenuEditorPaste:
			performCommand(COMMAND_PASTE);
			break;
		case R.id.MenuEditorUndo:
			performCommand(COMMAND_UNDO);
			break;
		case R.id.MenuEditorQuitWithoutSave:
			performCommand(COMMAND_QUIT);
			break;
		default:
			break;
		}

		return true;
	}

	private void processDone() {
		Intent intent = new Intent();
 		if (isModified) {
			intent.putExtra(KEY_STRING, mEditText.getText().toString());
			setResult(RESULT_OK, intent);
		} else {
			setResult(RESULT_CANCELED, intent);
		}
		finish();
	}

	private void doCut() {
		ClipboardManager cm = (ClipboardManager)getSystemService(CLIPBOARD_SERVICE);
		int start = mEditText.getSelectionStart();
		int end = mEditText.getSelectionEnd();
		if (start <= end) {
			cm.setText(mEditText.getText().subSequence(start, end));
			mEditText.getText().delete(start, end);
		} else {
			cm.setText(mEditText.getText().subSequence(end, start));
			mEditText.getText().delete(end, start);
		}
	}

	private void doCopy() {
		ClipboardManager cm = (ClipboardManager)getSystemService(CLIPBOARD_SERVICE);
		int start = mEditText.getSelectionStart();
		int end = mEditText.getSelectionEnd();
		if (start <= end) {
			cm.setText(mEditText.getText().subSequence(start, end));
		} else {
			cm.setText(mEditText.getText().subSequence(end, start));
		}
	}

	private void doPaste() {
		ClipboardManager cm = (ClipboardManager)getSystemService(CLIPBOARD_SERVICE);
		int start = mEditText.getSelectionStart();
		int end = mEditText.getSelectionEnd();
		if (start <= end) {
			mEditText.getText().delete(start, end);
		} else {
			mEditText.getText().delete(end, start);
		}
		insertStr(cm.getText());
	}

	private void insertStr(CharSequence str) {
		int pos = mEditText.getSelectionStart();
		mEditText.getText().insert(pos, str);
	}

	private void undo() {
		if (mEditHistory.empty()) return;

		EditHistoryItem item = mEditHistory.pop();
		isUndoing = true;
		mEditText.getText().replace(item.mStart, item.mStart+item.mCount, item.mBeforeText);
		isUndoing = false;
	}

	@Override
	protected Dialog onCreateDialog(int id) {
		Dialog dialog;
		switch(id) {
		case DIALOG_CONFIRM_EXIT:
			dialog = new AlertDialog.Builder(this)
				.setMessage(getString(R.string.label_confirm_back))
				.setPositiveButton(getString(R.string.label_YES), new DialogInterface.OnClickListener() {
					public void onClick(DialogInterface dialog, int whichButton) {
						setResult(RESULT_CANCELED);
						TextEditor.this.finish();
					}
				})
				.setNegativeButton(getString(R.string.label_NO), null)
				.create();
			break;
		default:
			dialog = null;
		}

		return dialog;
	}
}