package jp.deadend.noname.treenote;

import pl.polidea.treeview.InMemoryTreeStateManager;
import pl.polidea.treeview.InMemoryTreeStateEditInfo;
import pl.polidea.treeview.TreeNodeInfo;
import pl.polidea.treeview.TreeNoteBuilder;

import android.util.Log;
import android.util.SparseArray;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

class TreeNoteStateManager extends InMemoryTreeStateManager<Integer> implements Cloneable {
	private SparseArray<Stack<String>> mSubjects;
	private SparseArray<Stack<String>> mContents;
	private int mIDCounter;
	private TreeNote mActivity;
	private List<Integer> mBookmarks;

	public TreeNoteStateManager() {
		super(TreeNote.LEVEL_NUMBER);
		mSubjects = new SparseArray<Stack<String>>();
		mContents = new SparseArray<Stack<String>>();
		mIDCounter = 0;
		mActivity = null;
		mBookmarks = new ArrayList<Integer>();
	}

	private TreeNoteStateManager(int levelNumber, int idCounter) {
		super(levelNumber);
		mSubjects = new SparseArray<Stack<String>>();
		mContents = new SparseArray<Stack<String>>();
		mIDCounter = idCounter;
		mActivity = null;
		mBookmarks = new ArrayList<Integer>();
	}

	void registerActivity(TreeNote activity) {
		mActivity = activity;
	}

	void unregisterActivity() {
		mActivity = null;
	}

	@Override
	protected TreeNoteStateManager clone() {
		try {
			return (TreeNoteStateManager)super.clone();
		} catch (CloneNotSupportedException e) {
			Log.e("TreeNote", "clone() error");
			return null;
		}
	}

	@Override
	public void performEdit(InMemoryTreeStateEditInfo<Integer> editInfo) {
		if (editInfo instanceof TreeNoteEditInfo) {
			TreeNoteEditInfo ed = (TreeNoteEditInfo)editInfo;
			if (ed.getTextType() == TreeNoteEditInfo.TextEditType.SUBJECT) {
				mSubjects.get(ed.getId()).push(ed.getString());
			} else if (ed.getTextType() == TreeNoteEditInfo.TextEditType.CONTENT) {
				mContents.get(ed.getId()).push(ed.getString());
				mActivity.refreshContent();
			} else if (ed.getTextType() == TreeNoteEditInfo.TextEditType.ADD_NEW) {
				int id = ed.getId().intValue();
				if (mSubjects.get(id) == null) {
					Stack<String> stack_subj = new Stack<String>();
					stack_subj.push(ed.getString());
					Stack<String> stack_cont = new Stack<String>();
					stack_cont.push("");

					mSubjects.append(id, stack_subj);
					mContents.append(id, stack_cont);
				}
			}
		} else {
			super.performEdit(editInfo);
		}
	}

	@Override
	public void undoEdit(InMemoryTreeStateEditInfo<Integer> editInfo) {
		if (editInfo instanceof TreeNoteEditInfo) {
			TreeNoteEditInfo ed = (TreeNoteEditInfo)editInfo;
			if (ed.getTextType() == TreeNoteEditInfo.TextEditType.SUBJECT) {
				mSubjects.get(ed.getId()).pop();
			} else if (ed.getTextType() == TreeNoteEditInfo.TextEditType.CONTENT) {
				mContents.get(ed.getId()).pop();
				mActivity.refreshContent();
			} else if (ed.getTextType() == TreeNoteEditInfo.TextEditType.ADD_NEW) {
				// do nothing
			}
		} else {
			super.undoEdit(editInfo);
		}
	}

	int getRedoNum() {
		return mRedoBuffer.size();
	}

	void clearBookmarks() {
		mBookmarks.clear();
	}

	int getBookmarksSize() {
		return mBookmarks.size();
	}

	boolean isBookmarked(Integer id) {
		return mBookmarks.contains(id);
	}

	boolean addBookmark(Integer id) {
		if (mBookmarks.contains(id)) return false;
		mBookmarks.add(id);
		return true;
	}

	boolean removeBookmark(Integer id) {
		if (mBookmarks.contains(id)) return false;
		mBookmarks.remove(id);
		return true;
	}

	List<Integer> getBookmarks() {
		List<Integer> stub = new ArrayList<Integer>();
		for (Integer id: mBookmarks) {
			if (!isInTree(id)) {stub.add(id);}
		}
		mBookmarks.removeAll(stub);

		return mBookmarks;
	}

	boolean saveBookmarks(File file) {
		try {
			BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file, false), "UTF-8"), 1024);

			List<Integer> idlist = getIdList();
			for (int i=0; i<mBookmarks.size(); i++) {
				int idx = idlist.indexOf(mBookmarks.get(i));
				if (idx != -1) {
					bw.write(String.valueOf(idx));
					bw.write("\n");
				}
			}

			bw.close();
		} catch (IOException e) {
			Log.e("TreeNote", "saveBookmarks(): File IO error");
			return false;
		}

		return true;
	}

	static File getBookmarksFile(File contextDir, String fileName) {
		return new File(contextDir, fileName.replace('/', '_') + ".bkm");
	}

	ArrayList<ArrayList<InMemoryTreeStateEditInfo<Integer>>> convertUndoRedoBufferInHoisted(Integer rootId) {
		ArrayList<ArrayList<InMemoryTreeStateEditInfo<Integer>>> ret = new ArrayList<ArrayList<InMemoryTreeStateEditInfo<Integer>>>();

		for (List<InMemoryTreeStateEditInfo<Integer>> cmd : mUndoBuffer) {
			ArrayList<InMemoryTreeStateEditInfo<Integer>> newCmd = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
			for (InMemoryTreeStateEditInfo<Integer> edit : cmd) {
				InMemoryTreeStateEditInfo<Integer> clone = edit.createClone();
				if (rootId != null) {clone.convertRootInHoisted(rootId);}
				newCmd.add(clone);
			}
			ret.add(newCmd);
		}
		int undoLastIndex = ret.size();
		for (List<InMemoryTreeStateEditInfo<Integer>> cmd : mRedoBuffer) {
			ArrayList<InMemoryTreeStateEditInfo<Integer>> newCmd = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
			for (InMemoryTreeStateEditInfo<Integer> edit : cmd) {
				InMemoryTreeStateEditInfo<Integer> clone = edit.createClone();
				if (rootId != null) {clone.convertRootInHoisted(rootId);}
				newCmd.add(clone);
			}
			ret.add(undoLastIndex, newCmd);
		}

		return ret;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_addAfterChild(String subj, Integer afterChild) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		Integer id = Integer.valueOf(getNextIDCounter());

		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.ADD_NEW, id, subj));
		command.addAll(editCommand_addAfter(id, afterChild));

		return command;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_addAfterChild(String subj, String cont, Integer afterChild) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		Integer id = Integer.valueOf(getNextIDCounter());

		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.ADD_NEW, id, subj));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, id, cont));
		command.addAll(editCommand_addAfter(id, afterChild));

		return command;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_changeSubject(Integer id, String text) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();

		if (!containsID(id) || text == null) return null;
		if (text.equals(mSubjects.get(id).peek())) return null;

		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.SUBJECT, id, text));
		return command;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_changeContent(Integer id, String text) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList< InMemoryTreeStateEditInfo<Integer>>();

		if (!containsID(id) || text == null) return null;
		if (text.equals(mContents.get(id).peek())) return null;

		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, id, text));
		return command;
	}

	private void addEdit_copyNode(List<InMemoryTreeStateEditInfo<Integer>> command, Integer id, Integer newParent, int position) {
		mIDCounter++;
		Integer newId = Integer.valueOf(mIDCounter);
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.ADD_NEW, newId, mSubjects.get(id).peek()));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, newId, mContents.get(id).peek()));
		InMemoryTreeStateEditInfo<Integer> edit = new InMemoryTreeStateEditInfo<Integer>();
		edit.setAddChild(newParent, newId, position);
		command.add(edit);

		for (Integer child : getChildren(id)) {
			addEdit_copyNode(command, child, newId, -1);
		}
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_mergeToPrev(Integer id) {
		Integer prevId = getPreviousSibling(id);
		if (prevId == null) return null;

		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		String newSubj = mSubjects.get(prevId).peek() + mSubjects.get(id).peek();
		String newCont = mContents.get(prevId).peek() + mContents.get(id).peek();
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.SUBJECT, prevId, newSubj));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, prevId, newCont));

		for (Integer child : getChildren(id)) {
			addEdit_copyNode(command, child, prevId, -1);
		}

		command.addAll(editCommand_removeNodeRecursively(id));

		return command;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_mergeToNext(Integer id) {
		Integer nextId = getNextSibling(id);
		if (nextId == null) return null;

		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		String newSubj = mSubjects.get(id).peek() + mSubjects.get(nextId).peek();
		String newCont = mContents.get(id).peek() + mContents.get(nextId).peek();
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.SUBJECT, nextId, newSubj));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, nextId, newCont));

		List<Integer> children = getChildren(id);
		Collections.reverse(children);
		for (Integer child : children) {
			addEdit_copyNode(command, child, nextId, 0);
		}

		command.addAll(editCommand_removeNodeRecursively(id));

		return command;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_split(Integer id, int subj_from, int subj_to, int cont_from, int cont_to, String default_subj) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		String subj = mSubjects.get(id).peek();
		String cont = mContents.get(id).peek();
		String new_subj = subj.substring(subj_from, subj_to);
		String new_cont = cont.substring(cont_from, cont_to);
		String old_subj = subj.substring(0, subj_from) + subj.substring(subj_to);
		String old_cont = cont.substring(0, cont_from) + cont.substring(cont_to);

		if (new_subj.length() == 0) {
			command.addAll(editCommand_addAfterChild(default_subj, new_cont, id));
		} else {
			command.addAll(editCommand_addAfterChild(new_subj, new_cont, id));
		}
		if (!subj.equals(old_subj)) {
			if (old_subj.length() == 0) {
				command.addAll(editCommand_changeSubject(id, default_subj));
			} else {
				command.addAll(editCommand_changeSubject(id, old_subj));
			}
		}
		if (!cont.equals(old_cont)) {
			command.addAll(editCommand_changeContent(id, old_cont));
		}

		return command;
	}

	List<Integer> search(String str) {
		List<Integer> result = new ArrayList<Integer>();
		if (str == null) return result;
		if (str.length() == 0) return result;

		for (int i=0; i < mSubjects.size(); i++) {
			if (mSubjects.valueAt(i).peek().contains(str)) {
				int id = mSubjects.keyAt(i);
				if (!result.contains(id)) {
					result.add(id);
				}
			}
		}
		for (int i=0; i < mContents.size(); i++) {
			if (mContents.valueAt(i).peek().contains(str)) {
				int id = mContents.keyAt(i);
				if (!result.contains(id)) {
					result.add(id);
				}
			}
		}

		return result;
	}

	ArrayList<TreeNoteNodeInfo> listSearchedNodes(String str) {
		ArrayList<TreeNoteNodeInfo> result = new ArrayList<TreeNoteNodeInfo>();
		List<Integer> ids = search(str);
		Trace.d("listSearchedNodes: " + str + ", " + ids);

		for (Integer id : ids) {
			TreeNoteNodeInfo node = new TreeNoteNodeInfo();
			node.id = id;
			node.level = getLevel(id);
			node.subject = getSubject(id);
			node.content = getContent(id);
			node.hierarchyDesc = Arrays.asList(getHierarchyDescription(id)).toString();
			result.add(node);
		}

		return result;
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_replaceAll(String target, String replacement) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();

		for (int i=0; i < mSubjects.size(); i++) {
			String old_subj = mSubjects.valueAt(i).peek();
			String subj = old_subj.replaceAll(target, replacement);
			if (!subj.equals(old_subj)) {
				command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.SUBJECT, mSubjects.keyAt(i), subj));
			}

			String old_cont = mContents.valueAt(i).peek();
			String cont = old_cont.replaceAll(target, replacement);
			if (!cont.equals(old_cont)) {
				command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, mContents.keyAt(i), cont));
			}
		}

		if (command.isEmpty()) {return null;}
		return command;
	}

	String getSubject(Integer id) {
		return mSubjects.get(id).peek();
	}

	String getContent(Integer id) {
		return mContents.get(id).peek();
	}

	boolean containsID(Integer id) {
		if ((mSubjects.get(id) != null)|| (mContents.get(id) != null)) {
			return true;
		}
		return false;
	}

	int getNextIDCounter() {
		while (containsID(mIDCounter)) {
			mIDCounter++;
		}

		return mIDCounter;
	}

	Integer getLastID() {
		return Integer.valueOf(mIDCounter);
	}

	private void readChildrenToClipboard(Integer id, TreeNoteClipboardNode parentNode) {
		TreeNoteClipboardNode node = parentNode.add(mSubjects.get(id).peek(), mContents.get(id).peek());

		List<Integer> children = getChildren(id);
		for (Integer child: children) {
			readChildrenToClipboard(child, node);
		}
	}

	TreeNoteClipboardNode readToClipboard(Integer id) {
		if (id == null) return null;
		TreeNoteClipboardNode clipboard = new TreeNoteClipboardNode(mSubjects.get(id).peek(), mContents.get(id).peek());

		List<Integer> children = getChildren(id);
		for (Integer child: children) {
			readChildrenToClipboard(child, clipboard);
		}

		return clipboard;
	}

	private void addEdit_pasteClipChild(List<InMemoryTreeStateEditInfo<Integer>> command, TreeNoteClipboardNode clip, Integer parent) {
		mIDCounter++;
		Integer id = Integer.valueOf(mIDCounter);
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.ADD_NEW, id, clip.getSubject()));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, id, clip.getContent()));
		command.addAll(editCommand_addLast(id, parent));

		for (TreeNoteClipboardNode child : clip.getChildren()) {
			addEdit_pasteClipChild(command, child, id);
		}
	}

	List<InMemoryTreeStateEditInfo<Integer>> editCommand_pasteFromClipboard(TreeNoteClipboardNode clip, Integer beforeChild) {
		List<InMemoryTreeStateEditInfo<Integer>> command = new ArrayList<InMemoryTreeStateEditInfo<Integer>>();
		Integer id = Integer.valueOf(getNextIDCounter());
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.ADD_NEW, id, clip.getSubject()));
		command.add(new TreeNoteEditInfo(TreeNoteEditInfo.TextEditType.CONTENT, id, clip.getContent()));
		command.addAll(editCommand_addBefore(id, beforeChild));

		for (TreeNoteClipboardNode child : clip.getChildren()) {
			addEdit_pasteClipChild(command, child, id);
		}

		return command;
	}

	int getPosition(Integer id) {
		return getVisibleList().indexOf(id);
	}

	boolean writeNode(Writer w, Integer id, String indentStr, String newline) {
		if (id != null) {
			try {
				int level = getLevel(id);
				for (int i=0; i<(level+1); i++) {
					w.write(indentStr);
				}
				//~ w.write(" ");
				w.write(getSubject(id));
				w.write(newline);
				String content = getContent(id).trim().replace("\n", newline);
				if (content.length() > 0) {
					w.write(content);
					w.write(newline);
				}
			} catch (IOException e) {
				Log.e("TreeNote", "writeNode(): File IO error  id=" + id);
				return false;
			}
		}
		for (Integer child : getChildren(id)) {
			writeNode(w, child, indentStr, newline);
		}

		return true;
	}

	boolean saveFile(File file, String indentStr, String encoding, String newline) {
		try {
			BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file, false), encoding), 1024);
			if (!writeNode(bw, null, indentStr, newline)) return false;

			bw.flush();
			bw.close();
		} catch (IOException e) {
			Log.e("TreeNote", "saveFile(): File IO error");
			return false;
		}

		return true;
	}

	String listExpandedNodes() {
		StringBuilder sb = new StringBuilder();

		List<Integer> idlist = getIdList();
		List<Integer> expanded_list = new ArrayList<Integer>();
		for (Integer id : idlist) {
			if (id == null) continue;

			TreeNodeInfo<Integer> node = getNodeInfo(id);
			if (node.isExpanded()) {
				expanded_list.add(id);
			}
		}

		Collections.sort(expanded_list, new Comparator<Integer>(){
			public int compare(Integer id1, Integer id2) {
				return getLevel(id1) - getLevel(id2);
			}
		});

		for (Integer id : expanded_list) {
			Integer[] desc = getHierarchyDescription(id);
			for (Integer num : desc) {
				sb.append(num.toString());
				sb.append('-');
			}
			sb.append(',');
		}

		return sb.toString();
	}

	void expandAsSaved(Integer id, String state) {
		Trace.d("expandAsSaved(): " + state);
		if (state.length() == 0) {return;}

		String[] nodes = state.split(",");
		for (String node_str : nodes) {
			Integer node_to_expand = null; // root
			String[] h_desc = node_str.split("-");
			for (String child_pos : h_desc) {
				try {
					node_to_expand = getChildAt(node_to_expand, Integer.parseInt(child_pos));
				} catch (NumberFormatException e) {
				}
			}

			expandDirectChildren(node_to_expand);
		}
	}

	String getSubjectsFromRoot(Integer id) {
		if (mSubjects.get(id) == null) return "";
		StringBuilder sb = new StringBuilder();

		Integer currentID = id;
		while (currentID != null) {
			sb.insert(0, " - " + getSubject(currentID));
			currentID = getParent(currentID);
		}
		sb.delete(0, 3);

		return sb.toString();
	}

	void addToListBelowNode(List<TreeNoteNodeInfo> result, Integer id, int rootLevel) {
		if (!isInTree(id)) return;

		List<Integer> children = getChildren(id);
		for (Integer child : children) {
			TreeNoteNodeInfo node = new TreeNoteNodeInfo();
			node.id = child;
			node.level = getLevel(child) - (rootLevel+1);
			node.subject = getSubject(child);
			node.content = getContent(child);
			result.add(node);
			addToListBelowNode(result, child, rootLevel);
		}
	}

	private void setNewText(String subj, String cont, Integer id) {
		Stack<String> stack_subj = new Stack<String>();
		stack_subj.push(subj);
		Stack<String> stack_cont = new Stack<String>();
		stack_cont.push(cont);

		mSubjects.append(id, stack_subj);
		mContents.append(id, stack_cont);
	}

	static TreeNoteStateManager buildFromNodes(List<TreeNoteNodeInfo> nodes, int rootLevel, int idCounter) {
		if (nodes == null) {return null;}

		TreeNoteStateManager manager = new TreeNoteStateManager(TreeNote.LEVEL_NUMBER - (rootLevel+1), idCounter);
		if (nodes.size() == 0) {return manager;}
		TreeNoteBuilder<Integer> builder = new TreeNoteBuilder<Integer>(manager);

		for (TreeNoteNodeInfo node : nodes) {
			builder.sequentiallyAddNextNode(node.id, node.level);
			manager.setNewText(node.subject, node.content, node.id);
		}

		return manager;
	}

	private int forceAddTexts(String subj, String cont) {
		while (containsID(mIDCounter)) {
			mIDCounter++;
		}

		Stack<String> stack_subj = new Stack<String>();
		stack_subj.push(subj);
		Stack<String> stack_cont = new Stack<String>();
		stack_cont.push(cont);

		mSubjects.append(mIDCounter, stack_subj);
		mContents.append(mIDCounter, stack_cont);

		return mIDCounter;
	}

	private static List<Integer> loadBookmarks(File file) {
		List<Integer> result = new ArrayList<Integer>();
		if (!file.canRead()) {return result;}

		try {
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), "UTF-8"), 1024);
			String text;

			while((text = br.readLine()) != null) {
				Integer line = Integer.valueOf(Integer.parseInt(text));
				if (!result.contains(line)) {result.add(line);}
			}

			br.close();
		} catch (IOException e) {
			Log.e("TreeNote", "loadBookmarks(): File IO error");
			result.clear();
		} catch (NumberFormatException e) {
			Log.e("TreeNote", "loadBookmarks(): Wrong bookmark file");
			result.clear();
		}

		return result;
	}

	static TreeNoteStateManager loadFile(File file, String indentStr, String encoding, File bookmarkFile) {
		if (!file.canRead()) {return null;}

		TreeNoteStateManager manager = new TreeNoteStateManager();
		try {
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), encoding), 1024);
			String text = null;
			String subj = null;
			String stars = null;
			Integer id = 0;
			StringBuilder cont = null;
			int lineNo = 0;
//			Pattern pattern = Pattern.compile("^(\\*+)\\s*(.*)");
			Pattern pattern = Pattern.compile("^(\\" + indentStr + "+)\\s*(.*)");
			Matcher matcher = null;
			do {
				text = br.readLine();
				if (text == null) {
					br.close();
					return null;
				}
				matcher = pattern.matcher(text);
			} while (!matcher.find());

			List<Integer> bookmarks = loadBookmarks(bookmarkFile);
			TreeNoteBuilder<Integer> builder = new TreeNoteBuilder<Integer>(manager);
			while (true) {
				stars = matcher.group(1);
				subj = matcher.group(2);
				text = br.readLine();
				if (text == null) {
					id = manager.forceAddTexts(subj, "");
					builder.sequentiallyAddNextNode(id, stars.length()-1);
					if (bookmarks.contains(Integer.valueOf(lineNo))) {
						manager.addBookmark(id);
					}
					br.close();
					return manager;
				}
				cont = new StringBuilder();
				matcher = pattern.matcher(text);
				while (!matcher.find()) {
					cont.append(text);
					cont.append("\n");
					text = br.readLine();
					if (text == null) {
						id = manager.forceAddTexts(subj, cont.toString());
						builder.sequentiallyAddNextNode(id, stars.length()-1);
						br.close();
						return manager;
					}
					matcher = pattern.matcher(text);
				}
				id = manager.forceAddTexts(subj, cont.toString());
				builder.sequentiallyAddNextNode(id, stars.length()-1);
				if (bookmarks.contains(Integer.valueOf(lineNo))) {
					manager.addBookmark(id);
				}
				lineNo++;
			}
		} catch (IOException e) {
			Log.e("TreeNote", "loadFile(): File IO error");
			return null;
		}
	}
}
