package pl.polidea.treeview;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import android.database.DataSetObserver;
import android.util.Log;

import jp.deadend.noname.treenote.Trace;

/**
 * In-memory manager of tree state.
 *
 * @param <T>
 *            type of identifier
 */
public class InMemoryTreeStateManager<T extends Serializable> implements TreeStateManager<T> {
	protected Stack<List<InMemoryTreeStateEditInfo<T>>> mUndoBuffer;
	protected Stack<List<InMemoryTreeStateEditInfo<T>>> mRedoBuffer;

    private static final String TAG = InMemoryTreeStateManager.class
            .getSimpleName();
    private static final long serialVersionUID = 1L;
    private final Map<T, InMemoryTreeNode<T>> allNodes = new HashMap<T, InMemoryTreeNode<T>>();
    private final InMemoryTreeNode<T> topSentinel = new InMemoryTreeNode<T>(
            null, null, -1, true);
    private transient List<T> visibleListCache = null; // lasy initialised
    private transient List<T> unmodifiableVisibleList = null;
    private boolean visibleByDefault = true;
    private final transient Set<DataSetObserver> observers = new HashSet<DataSetObserver>();
    private final int numberOfLevels;

	public InMemoryTreeStateManager(int nol) {
		this.numberOfLevels = nol;
		this.mUndoBuffer = new Stack<List<InMemoryTreeStateEditInfo<T>>>();
		this.mRedoBuffer = new Stack<List<InMemoryTreeStateEditInfo<T>>>();
	}

    private synchronized void internalDataSetChanged() {
        visibleListCache = null;
        unmodifiableVisibleList = null;
        for (final DataSetObserver observer : observers) {
            observer.onChanged();
        }
    }

    /**
     * If true new nodes are visible by default.
     *
     * @param visibleByDefault
     *            if true, then newly added nodes are expanded by default
     */
    public void setVisibleByDefault(final boolean visibleByDefault) {
        this.visibleByDefault = visibleByDefault;
    }

    private InMemoryTreeNode<T> getNodeFromTreeOrThrow(final T id) {
        if (id == null) {
            throw new NodeNotInTreeException("(null)");
        }
        final InMemoryTreeNode<T> node = allNodes.get(id);
        if (node == null) {
            throw new NodeNotInTreeException(id.toString());
        }
        return node;
    }

    private InMemoryTreeNode<T> getNodeFromTreeOrThrowAllowRoot(final T id) {
        if (id == null) {
            return topSentinel;
        }
        return getNodeFromTreeOrThrow(id);
    }

    private void expectNodeNotInTreeYet(final T id) {
        final InMemoryTreeNode<T> node = allNodes.get(id);
        if (node != null) {
            throw new NodeAlreadyInTreeException(id.toString(), node.toString());
        }
    }

    @Override
    public synchronized TreeNodeInfo<T> getNodeInfo(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrow(id);
        final List<InMemoryTreeNode<T>> children = node.getChildren();
        boolean expanded = false;
        if (!children.isEmpty() && children.get(0).isVisible()) {
            expanded = true;
        }
        return new TreeNodeInfo<T>(id, node.getLevel(), !children.isEmpty(),
                node.isVisible(), expanded);
    }

    @Override
    public synchronized List<T> getChildren(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        return node.getChildIdList();
    }

    @Override
    public synchronized T getParent(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        return node.getParent();
    }

    private boolean getChildrenVisibility(final InMemoryTreeNode<T> node) {
        boolean visibility;
        final List<InMemoryTreeNode<T>> children = node.getChildren();
        if (children.isEmpty()) {
            visibility = visibleByDefault;
        } else {
            visibility = children.get(0).isVisible();
        }
        return visibility;
    }

    @Override
    public synchronized void addBeforeChild(final T parent, final T newChild,
            final T beforeChild) {
        expectNodeNotInTreeYet(newChild);
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(parent);
        final boolean visibility = getChildrenVisibility(node);
        // top nodes are always expanded.
        if (beforeChild == null) {
            final InMemoryTreeNode<T> added = node.add(0, newChild, visibility);
            allNodes.put(newChild, added);
        } else {
            final int index = node.indexOf(beforeChild);
            final InMemoryTreeNode<T> added = node.add(index == -1 ? 0 : index,
                    newChild, visibility);
            allNodes.put(newChild, added);
        }
        if (visibility) {
            internalDataSetChanged();
        }
    }

    @Override
    public synchronized void addAfterChild(final T parent, final T newChild,
            final T afterChild) {
        expectNodeNotInTreeYet(newChild);
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(parent);
        final boolean visibility = getChildrenVisibility(node);
        if (afterChild == null) {
            final InMemoryTreeNode<T> added = node.add(
                    node.getChildrenListSize(), newChild, visibility);
            allNodes.put(newChild, added);
        } else {
            final int index = node.indexOf(afterChild);
            final InMemoryTreeNode<T> added = node.add(
                    index == -1 ? node.getChildrenListSize() : index + 1, newChild,
                    visibility);
            allNodes.put(newChild, added);
        }
        if (visibility) {
            internalDataSetChanged();
        }
    }

    @Override
    public synchronized void removeNodeRecursively(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        final boolean visibleNodeChanged = removeNodeRecursively(node);
        final T parent = node.getParent();
        final InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
        parentNode.removeChild(id);
        if (visibleNodeChanged) {
            internalDataSetChanged();
        }
    }

    private boolean removeNodeRecursively(final InMemoryTreeNode<T> node) {
        boolean visibleNodeChanged = false;
        for (final InMemoryTreeNode<T> child : node.getChildren()) {
            if (removeNodeRecursively(child)) {
                visibleNodeChanged = true;
            }
        }
        node.clearChildren();
        if (node.getId() != null) {
            allNodes.remove(node.getId());
            if (node.isVisible()) {
                visibleNodeChanged = true;
            }
        }
        return visibleNodeChanged;
    }

    private void setChildrenVisibility(final InMemoryTreeNode<T> node,
            final boolean visible, final boolean recursive) {
        for (final InMemoryTreeNode<T> child : node.getChildren()) {
            child.setVisible(visible);
            if (recursive) {
                setChildrenVisibility(child, visible, true);
            }
        }
    }

    @Override
    public synchronized void expandDirectChildren(final T id) {
        //~ Log.d(TAG, "Expanding direct children of " + id);
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        setChildrenVisibility(node, true, false);
        internalDataSetChanged();
    }

    @Override
    public synchronized void expandEverythingBelow(final T id) {
        //~ Log.d(TAG, "Expanding all children below " + id);
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        setChildrenVisibility(node, true, true);
        internalDataSetChanged();
    }

    @Override
    public synchronized void collapseChildren(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        if (node == topSentinel) {
            for (final InMemoryTreeNode<T> n : topSentinel.getChildren()) {
                setChildrenVisibility(n, false, true);
            }
        } else {
            setChildrenVisibility(node, false, true);
        }
        internalDataSetChanged();
    }

    @Override
    public synchronized T getNextSibling(final T id) {
        final T parent = getParent(id);
        final InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
        boolean returnNext = false;
        for (final InMemoryTreeNode<T> child : parentNode.getChildren()) {
            if (returnNext) {
                return child.getId();
            }
            if (child.getId().equals(id)) {
                returnNext = true;
            }
        }
        return null;
    }

    @Override
    public synchronized T getPreviousSibling(final T id) {
        final T parent = getParent(id);
        final InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
        T previousSibling = null;
        for (final InMemoryTreeNode<T> child : parentNode.getChildren()) {
            if (child.getId().equals(id)) {
                return previousSibling;
            }
			previousSibling = child.getId();
        }
        return null;
    }

    @Override
    public synchronized boolean isInTree(final T id) {
        return allNodes.containsKey(id);
    }

    @Override
    public synchronized int getVisibleCount() {
        return getVisibleList().size();
    }

    @Override
    public synchronized List<T> getVisibleList() {
        T currentId = null;
        if (visibleListCache == null) {
            visibleListCache = new ArrayList<T>(allNodes.size());
            do {
                currentId = getNextVisible(currentId);
                if (currentId == null) {
                    break;
                } else {
                    visibleListCache.add(currentId);
                }
            } while (true);
        }
        if (unmodifiableVisibleList == null) {
            unmodifiableVisibleList = Collections
                    .unmodifiableList(visibleListCache);
        }
        return unmodifiableVisibleList;
    }

    public synchronized T getNextVisible(final T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        if (!node.isVisible()) {
            return null;
        }
        final List<InMemoryTreeNode<T>> children = node.getChildren();
        if (!children.isEmpty()) {
            final InMemoryTreeNode<T> firstChild = children.get(0);
            if (firstChild.isVisible()) {
                return firstChild.getId();
            }
        }
        final T sibl = getNextSibling(id);
        if (sibl != null) {
            return sibl;
        }
        T parent = node.getParent();
        do {
            if (parent == null) {
                return null;
            }
            final T parentSibling = getNextSibling(parent);
            if (parentSibling != null) {
                return parentSibling;
            }
            parent = getNodeFromTreeOrThrow(parent).getParent();
        } while (true);
    }

    @Override
    public synchronized void registerDataSetObserver(
            final DataSetObserver observer) {
        observers.add(observer);
    }

    @Override
    public synchronized void unregisterDataSetObserver(
            final DataSetObserver observer) {
        observers.remove(observer);
    }

    @Override
    public int getLevel(final T id) {
        return getNodeFromTreeOrThrow(id).getLevel();
    }

    @Override
    public Integer[] getHierarchyDescription(final T id) {
        final int level = getLevel(id);
        final Integer[] hierarchy = new Integer[level + 1];
        int currentLevel = level;
        T currentId = id;
        T parent = getParent(currentId);
        while (currentLevel >= 0) {
            hierarchy[currentLevel--] = getChildren(parent).indexOf(currentId);
            currentId = parent;
            parent = getParent(parent);
        }
        return hierarchy;
    }

    private void appendToSb(final StringBuilder sb, final T id) {
        if (id != null) {
            final TreeNodeInfo<T> node = getNodeInfo(id);
            final int indent = node.getLevel() * 4;
            final char[] indentString = new char[indent];
            Arrays.fill(indentString, ' ');
            sb.append(indentString);
            sb.append(node.toString());
sb.append(" parent=");
sb.append(getParent(id));
sb.append(" ");
            sb.append(Arrays.asList(getHierarchyDescription(id)).toString());
            sb.append("\n");
        }
        final List<T> children = getChildren(id);
        for (final T child : children) {
            appendToSb(sb, child);
        }
    }

    @Override
    public synchronized String toString() {
        final StringBuilder sb = new StringBuilder();
        appendToSb(sb, null);
        return sb.toString();
    }

    @Override
    public synchronized void clear() {
        allNodes.clear();
        topSentinel.clearChildren();
        internalDataSetChanged();
    }

    @Override
    public void refresh() {
        internalDataSetChanged();
    }



    private void expectNodeInTree(final T id) {
        final InMemoryTreeNode<T> node = allNodes.get(id);
        if (node == null) {
            throw new NodeNotInTreeException(id.toString());
        }
    }

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_removeNodeRecursively(T id) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
        addEdit_removeNodeRecursively(command, id);

		return command;
	}

	private void addEdit_removeNodeRecursively(List<InMemoryTreeStateEditInfo<T>> command, T id) {
        final InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
		List<T> children = node.getChildIdList();
		int size = children.size();
		for (int i = size-1; i >= 0; i--) {
            addEdit_removeNodeRecursively(command, children.get(i));
		}
        if (id != null) {
			InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
			edit.setRemoveNode(id, getParent(id), getPosInParent(id));
			command.add(edit);
        }
	}

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_addAfter(T id, T afterChild) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
		edit.setAddChild(getParent(afterChild), id, getPosInParent(afterChild)+1);
		command.add(edit);

		return command;
	}

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_addBefore(T id, T beforeChild) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
		edit.setAddChild(getParent(beforeChild), id, getPosInParent(beforeChild));
		command.add(edit);

		return command;
	}

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_addLast(T id, T parent) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
		edit.setAddChild(parent, id, -1);
		command.add(edit);

		return command;
	}

	void raiseLevelSubtree(T id) {
        expectNodeInTree(id);
        List<T> children = getChildren(id);

		for (T child : children) {
			raiseLevelSubtree(child);
		}
		getNodeFromTreeOrThrow(id).raiseLevel();
	}

	void lowerLevelSubtree(T id) {
        expectNodeInTree(id);
        List<T> children = getChildren(id);

		for (T child : children) {
			lowerLevelSubtree(child);
		}
		getNodeFromTreeOrThrow(id).lowerLevel();
	}

    private synchronized List<T> getSiblingsAfter(final T id) {
		List<T> result = new ArrayList<T>();
        final T parent = getParent(id);
        final InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
        boolean isAfterId = false;
        for (final InMemoryTreeNode<T> child : parentNode.getChildren()) {
            if (isAfterId) {
				result.add(child.getId());
            }
            if (child.getId().equals(id)) {
                isAfterId = true;
            }
        }

        return result;
    }

	@Override
	public synchronized List<InMemoryTreeStateEditInfo<T>> editCommand_promote(T id) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		InMemoryTreeNode<T> node = null;
		try {
			node = getNodeFromTreeOrThrow(id);
		} catch (NodeNotInTreeException e) {
			return null;
		}
		if (node.getLevel() <= 0) return null;

		InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
		edit.setRaise(id);
		command.add(edit);

		List<T> newChildren = getSiblingsAfter(id);
		for (T child : newChildren) {
			if (node.getLevel()-1 < getLevel(child)) {
				InMemoryTreeStateEditInfo<T> edit2 = new InMemoryTreeStateEditInfo<T>();
				edit2.setChangeParent(child, getParent(child), id, getPosInParent(child), -1);
				command.add(edit2);
			}
		}

		T oldParent = getParent(id);
		if (oldParent != null) {
			T newParent = getParent(oldParent);
			if (node.getLevel()-1 <= getLevel(oldParent)) {
				InMemoryTreeStateEditInfo<T> edit3 = new InMemoryTreeStateEditInfo<T>();
				edit3.setChangeParent(id, oldParent, newParent, getPosInParent(id), getPosInParent(oldParent)+1);
				command.add(edit3);
			}
		}

		return command;
	}

	private int getChildrenMaxLevel(T id) {
		int ret = getLevel(id);

		for (T child : getChildren(id)) {
			if (getChildrenMaxLevel(child) > ret) {
				ret = getLevel(child);
			}
		}

		return ret;
	}

	@Override
	public synchronized List<InMemoryTreeStateEditInfo<T>> editCommand_demote(T id) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		InMemoryTreeNode<T> node = null;
		try {
			node = getNodeFromTreeOrThrow(id);
		} catch (NodeNotInTreeException e) {
			return null;
		}
		if (getChildrenMaxLevel(id) >= numberOfLevels-2) return null;

		InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
		edit.setLower(id);
		command.add(edit);

		T oldParent = getParent(id);
		T newParent = getPreviousSibling(id);
		if (newParent != null) {
			if (node.getLevel()+1 > getLevel(newParent)) {
				InMemoryTreeStateEditInfo<T> edit2 = new InMemoryTreeStateEditInfo<T>();
				edit2.setChangeParent(id, oldParent, newParent, getPosInParent(id), -1);
				command.add(edit2);
			}
		}

		return command;
	}

	private T getNextSiblingOrSubtree(T id) {
		if (id == null) return null;

		T parent = getParent(id);
		T next = getNextSibling(id);
		if (next != null) return next;

		return getNextSiblingOrSubtree(parent);
	}

	private T getPreviousSiblingOrSubtree(T id) {
		if (id == null) return null;

		T parent = getParent(id);
		T previous = getPreviousSibling(id);
		if (previous != null) return previous;

		return getPreviousSiblingOrSubtree(parent);
	}

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_moveUp(T id) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		if (id == null) return null;
		expectNodeInTree(id);

		T parent = getParent(id);
		InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
		int oldIndex = parentNode.indexOf(id);
		if (oldIndex != 0) {
			InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
			edit.setMoveUp(id);
			command.add(edit);
		} else {
			T newParent = getPreviousSiblingOrSubtree(id);
			if (newParent != null) {
				InMemoryTreeStateEditInfo<T> edit2 = new InMemoryTreeStateEditInfo<T>();
				edit2.setChangeParent(id, parent, newParent, oldIndex, -1);
				command.add(edit2);
			} else {
				return null;
			}
		}

		return command;
	}

	@Override
	public List<InMemoryTreeStateEditInfo<T>> editCommand_moveDown(T id) {
		List<InMemoryTreeStateEditInfo<T>> command = new ArrayList<InMemoryTreeStateEditInfo<T>>();
		if (id == null) return null;
		expectNodeInTree(id);

		T parent = getParent(id);
		InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(parent);
		int oldIndex = parentNode.indexOf(id);
		if (oldIndex < parentNode.getChildrenListSize()-1) {
			InMemoryTreeStateEditInfo<T> edit = new InMemoryTreeStateEditInfo<T>();
			edit.setMoveDown(id);
			command.add(edit);
		} else {
			T newParent = getNextSiblingOrSubtree(id);
			if (newParent != null) {
				InMemoryTreeStateEditInfo<T> edit2 = new InMemoryTreeStateEditInfo<T>();
				edit2.setChangeParent(id, parent, newParent, oldIndex, 0);
				command.add(edit2);
			} else {
				return null;
			}
		}

		return command;
	}

	@Override
	public boolean commitCommand(List<InMemoryTreeStateEditInfo<T>> command) {
		if (command != null) {
			performCommand(command);
			mUndoBuffer.push(command);
			mRedoBuffer.clear();

			return true;
		}

		return false;
	}

	@Override
	public void performUndo() {
		if (mUndoBuffer.empty()) return;

		List<InMemoryTreeStateEditInfo<T>> command = mUndoBuffer.pop();
		undoCommand(command);
		mRedoBuffer.push(command);
	}

	@Override
	public void performRedo() {
		if (mRedoBuffer.empty()) return;

		List<InMemoryTreeStateEditInfo<T>> command = mRedoBuffer.pop();
		performCommand(command);
		mUndoBuffer.push(command);
	}

	@Override
	public void clearEditHistory() {
		mUndoBuffer.clear();
		mRedoBuffer.clear();
	}

	@Override
	public boolean isUndoable() {
		return (!mUndoBuffer.empty());
	}

	@Override
	public boolean isRedoable() {
		return (!mRedoBuffer.empty());
	}

	public void performEdit(InMemoryTreeStateEditInfo<T> editInfo) {
		switch (editInfo.getType()) {
		case ADD_CHILD: {
			T id = editInfo.getId1();
			InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrowAllowRoot(id);

			int index = editInfo.getIndex1();
			if (index < 0) {
				index = parentNode.getChildrenListSize();
			}

			expandToId(id);
			T childId = editInfo.getId2();
			InMemoryTreeNode<T> added = parentNode.add(index, childId, true);
			allNodes.put(childId, added);
			break; }
		case REMOVE_NODE: {
			T id = editInfo.getId1();
			T parentId = editInfo.getId2();
			expandToId(id);
			getNodeFromTreeOrThrowAllowRoot(parentId).removeChild(id);
			allNodes.remove(id);
			break; }
		case CHANGE_PARENT: {
			T id = editInfo.getId1();
			T oldParentId = editInfo.getId2();
			T newParentId = editInfo.getId3();
			expandToId(id);
			expandToId(newParentId);

			InMemoryTreeNode<T> node = getNodeFromTreeOrThrow(id);
			InMemoryTreeNode<T> oldParentNode = getNodeFromTreeOrThrowAllowRoot(oldParentId);
			InMemoryTreeNode<T> newParentNode = getNodeFromTreeOrThrowAllowRoot(newParentId);
			int newIndex = editInfo.getIndex2();
			if (newIndex < 0) {
				newParentNode.insertChild(node);
			} else {
				newParentNode.insertChild(newIndex, node);
			}
			oldParentNode.removeChild(id);
			break; }
		case MOVE_UP: {
			T id = editInfo.getId1();
			expandToId(id);
			getNodeFromTreeOrThrowAllowRoot(getParent(id)).moveUpChild(id);
			break; }
		case MOVE_DOWN: {
			T id = editInfo.getId1();
			expandToId(id);
			getNodeFromTreeOrThrowAllowRoot(getParent(id)).moveDownChild(id);
			break; }
		case RAISE: {
			T id = editInfo.getId1();
			expandToId(id);
			raiseLevelSubtree(id);
			break; }
		case LOWER: {
			T id = editInfo.getId1();
			expandToId(id);
			lowerLevelSubtree(id);
			break; }
		}
	}

	public void undoEdit(InMemoryTreeStateEditInfo<T> editInfo) {
		switch (editInfo.getType()) {
		case ADD_CHILD: {
			T id = editInfo.getId1();
			T childId = editInfo.getId2();
			expandToId(childId);
			getNodeFromTreeOrThrowAllowRoot(id).removeChild(childId);
			allNodes.remove(childId);
			break; }
		case REMOVE_NODE: {
			T id = editInfo.getId1();
			T parentId = editInfo.getId2();
			expandToId(parentId);
			InMemoryTreeNode<T> added = getNodeFromTreeOrThrowAllowRoot(parentId).add(editInfo.getIndex1(), id, true);
			allNodes.put(id, added);
			break; }
		case CHANGE_PARENT: {
			T id = editInfo.getId1();
			T oldParentId = editInfo.getId2();
			T newParentId = editInfo.getId3();
			expandToId(id);
			expandToId(oldParentId);

			InMemoryTreeNode<T> node = getNodeFromTreeOrThrow(id);
			InMemoryTreeNode<T> oldParentNode = getNodeFromTreeOrThrowAllowRoot(oldParentId);
			InMemoryTreeNode<T> newParentNode = getNodeFromTreeOrThrowAllowRoot(newParentId);
			oldParentNode.insertChild(editInfo.getIndex1(), node);
			newParentNode.removeChild(id);
			break; }
		case MOVE_UP: {
			T id = editInfo.getId1();
			expandToId(id);
			getNodeFromTreeOrThrowAllowRoot(getParent(id)).moveDownChild(id);
			break; }
		case MOVE_DOWN: {
			T id = editInfo.getId1();
			expandToId(id);
			getNodeFromTreeOrThrowAllowRoot(getParent(id)).moveUpChild(id);
			break; }
		case RAISE: {
			T id = editInfo.getId1();
			expandToId(id);
			lowerLevelSubtree(id);
			break; }
		case LOWER: {
			T id = editInfo.getId1();
			expandToId(id);
			raiseLevelSubtree(id);
			break; }
		}
	}

	private void performCommand(List<InMemoryTreeStateEditInfo<T>> command) {
		int size = command.size();
		for (int i = 0; i < size; i++) {
			performEdit(command.get(i));
		}
		internalDataSetChanged();
	}

	private void undoCommand(List<InMemoryTreeStateEditInfo<T>> command) {
		int size = command.size();
		for (int i = size-1; i >= 0; i--) {
			undoEdit(command.get(i));
		}
		internalDataSetChanged();
	}

    private void getIdWithChildren(List<T> list, final T id) {
        if (id != null) {
			list.add(id);
        }
        final List<T> children = getChildren(id);
        for (final T child : children) {
            getIdWithChildren(list, child);
        }
    }

    public List<T> getIdList() {
		List<T> result = new ArrayList<T>();
		getIdWithChildren(result, null);
		return result;
	}

	private void expandParent(InMemoryTreeNode<T> node) {
		if (node.isVisible()) return;

		final T parent = node.getParent();
		if (parent == null) return; // all nodes are collapsed?

		final InMemoryTreeNode<T> parentNode = getNodeFromTreeOrThrow(parent);
		if (parentNode.isVisible()) {
			setChildrenVisibility(parentNode, true, true);
			internalDataSetChanged();
		} else {
			expandParent(parentNode);
		}
	}

	public void expandToId(final T id) {
		if (id == null) return;

		final InMemoryTreeNode<T> node = getNodeFromTreeOrThrow(id);
		if (node.isVisible()) return;
		expandParent(node);
	}

	public T getChildAt(final T id, int pos) {
        InMemoryTreeNode<T> node = getNodeFromTreeOrThrowAllowRoot(id);
        if (pos >= node.getChildrenListSize()) return id;

        return node.getChildren().get(pos).getId();
	}

	public int getPosInParent(final T id) {
		return getNodeFromTreeOrThrowAllowRoot(getParent(id)).indexOf(id);
	}
}
